@@ -11,23 +11,23 @@ import requests
 from pygrocy.utils import parse_date, parse_float, parse_int
 from tzlocal import get_localzone
 
-DEFAULT_PORT_NUMBER=9192
+DEFAULT_PORT_NUMBER = 9192
+
 
 class ShoppingListItem(object):
     def __init__(self, parsed_json):
         self._id = parse_int(parsed_json.get('id'))
         self._product_id = parse_int(parsed_json.get('product_id', None))
-        self._note = parsed_json.get('note',None)
-        self._amount = parse_float(parsed_json.get('amount'),0)
+        self._note = parsed_json.get('note', None)
+        self._amount = parse_float(parsed_json.get('amount'), 0)
         self._row_created_timestamp = parse_date(parsed_json.get('row_created_timestamp', None))
         self._shopping_list_id = parse_int(parsed_json.get('shopping_list_id'))
         self._done = parse_int(parsed_json.get('done'))
 
-
     @property
     def id(self) -> int:
         return self._id
-    
+
     @property
     def product_id(self) -> int:
         return self._product_id
@@ -40,6 +40,7 @@ class ShoppingListItem(object):
     def amount(self) -> float:
         return self._amount
 
+
 class QuantityUnitData(object):
     def __init__(self, parsed_json):
         self._id = parse_int(parsed_json.get('id'))
@@ -55,15 +56,15 @@ class LocationData(object):
         self._name = parsed_json.get('name')
         self._description = parsed_json.get('description')
         self._row_created_timestamp = parse_date(parsed_json.get('row_created_timestamp'))
-        
+
     @property
     def id(self) -> int:
         return self._id
-        
+
     @property
     def name(self) -> str:
         return self._name
-        
+
     @property
     def description(self) -> str:
         return self._description
@@ -91,11 +92,10 @@ class ProductData(object):
         else:
             self._barcodes = barcodes_raw.split(",")
 
-
     @property
     def id(self) -> int:
         return self._id
-        
+
     @property
     def product_group_id(self) -> int:
         return self._product_group_id
@@ -111,16 +111,18 @@ class ProductData(object):
 
 class ChoreData(object):
     def __init__(self, parsed_json):
-        self._id = parse_int(parsed_json.get('id'))
-        self._name = parsed_json.get('name')
-
-    @property
-    def id(self) -> int:
-        return self._id
-
-    @property
-    def name(self) -> str:
-        return self._name
+        self.id = parse_int(parsed_json.get('id'))
+        self.name = parsed_json.get('name')
+        self.description = parsed_json.get('description')
+        self.period_type = parsed_json.get('period_type')
+        self.period_config = parsed_json.get('period_config')
+        self.period_days = parse_int(parsed_json.get('period_days'))
+        self.track_date_only = parsed_json.get('track_date_only')
+        self.rollover = parsed_json.get('rollover')
+        self.assignment_type = parsed_json.get('assignment_type')
+        self.assignment_config = parsed_json.get('assignment_config')
+        self.next_execution_assigned_to_user_id = parse_int('next_execution_assigned_to_user_id')
+        self.userfields = parsed_json.get('userfields')
 
 
 class UserDto(object):
@@ -172,14 +174,14 @@ class CurrentChoreResponse(object):
         return self._next_estimated_execution_time
 
 
-
 class CurrentStockResponse(object):
     def __init__(self, parsed_json):
         self._product_id = parse_int(parsed_json.get('product_id'))
         self._amount = parse_float(parsed_json.get('amount'))
         self._best_before_date = parse_date(parsed_json.get('best_before_date'))
         self._amount_opened = parse_float(parsed_json.get('amount_opened'))
-        
+        self._product = ProductData(parsed_json.get('product'))
+
     @property
     def product_id(self) -> int:
         return self._product_id
@@ -196,6 +198,11 @@ class CurrentStockResponse(object):
     def amount_opened(self) -> float:
         return self._amount_opened
 
+    @property
+    def product(self) -> ProductData:
+        return self._product
+
+
 class MissingProductResponse(object):
     def __init__(self, parsed_json):
         self._product_id = parse_int(parsed_json.get('id'))
@@ -219,6 +226,7 @@ class MissingProductResponse(object):
     def is_partly_in_stock(self) -> bool:
         return self._is_partly_in_stock
 
+
 class CurrentVolatilStockResponse(object):
     def __init__(self, parsed_json):
         self._expiring_products = [CurrentStockResponse(product) for product in parsed_json.get('expiring_products')]
@@ -238,7 +246,6 @@ class CurrentVolatilStockResponse(object):
         return self._missing_products
 
 
-
 class ProductDetailsResponse(object):
     def __init__(self, parsed_json):
         self._last_purchased = parse_date(parsed_json.get('last_purchased'))
@@ -288,6 +295,14 @@ class ChoreDetailsResponse(object):
     def __init__(self, parsed_json):
         self._chore = ChoreData(parsed_json.get('chore'))
         self._last_tracked = parse_date(parsed_json.get('last_tracked'))
+        self._next_estimated_execution_time = parse_date(parsed_json.get('next_estimated_execution_time'))
+        self._track_count = parse_int(parsed_json.get('track_count'))
+
+        next_user = parsed_json.get('next_execution_assigned_user')
+        if next_user is not None:
+            self._next_execution_assigned_user = UserDto(next_user)
+        else:
+            self._next_execution_assigned_user = None
 
         if self._last_tracked is None:
             self._last_done_by = None
@@ -306,6 +321,18 @@ class ChoreDetailsResponse(object):
     def last_tracked(self) -> datetime:
         return self._last_tracked
 
+    @property
+    def next_estimated_execution_time(self) -> datetime:
+        return self._next_estimated_execution_time
+
+    @property
+    def track_count(self) -> int:
+        return self._track_count
+
+    @property
+    def next_execution_assigned_user(self) -> UserDto:
+        return self._next_execution_assigned_user
+
 
 class TransactionType(Enum):
     PURCHASE = "purchase"
@@ -314,52 +341,90 @@ class TransactionType(Enum):
     PRODUCT_OPENED = "product-opened"
 
 
+class TaskResponse(object):
+    def __init__(self, parsed_json):
+        self.id = parse_int(parsed_json.get('id'))
+        self.name = parsed_json.get('name')
+        self.description = parsed_json.get('description')
+        self.due_date = parse_date(parsed_json.get('due_date'))
+        self.done = parse_int(parsed_json.get('done'))
+        self.done_timestamp = parse_date(parsed_json.get('done_timestamp'))
+        self.category_id = parse_int(parsed_json.get('category_id'))
+        self.assigned_to_user_id = parse_int(parsed_json.get('assigned_to_user_id'))
+        self.userfields = parsed_json.get('userfields')
+
+
 class GrocyApiClient(object):
-    def __init__(self, base_url, api_key, port: int = DEFAULT_PORT_NUMBER, verify_ssl = True):
+    def __init__(self, base_url, api_key, port: int = DEFAULT_PORT_NUMBER, verify_ssl=True):
         self._base_url = '{}:{}/api/'.format(base_url, port)
         self._api_key = api_key
         self._verify_ssl = verify_ssl
         if self._api_key == "demo_mode":
-            self._headers = { "accept": "application/json" }
+            self._headers = {"accept": "application/json"}
         else:
             self._headers = {
                 "accept": "application/json",
                 "GROCY-API-KEY": api_key
             }
 
+    def _do_get_request(self, end_url: str):
+        req_url = urljoin(self._base_url, end_url)
+        resp = requests.get(
+            req_url, verify=self._verify_ssl, headers=self._headers)
+        resp.raise_for_status()
+        if len(resp.content) > 0:
+            return resp.json()
+
+    def _do_post_request(self, end_url: str, data: dict):
+        req_url = urljoin(self._base_url, end_url)
+        resp = requests.post(
+            req_url, verify=self._verify_ssl,
+            headers=self._headers,
+            data=data)
+        resp.raise_for_status()
+        if len(resp.content) > 0:
+            return resp.json()
+
+    def _do_put_request(self, end_url: str, data):
+        req_url = urljoin(self._base_url, end_url)
+        up_header = self._headers.copy()
+        up_header['accept'] = '*/*'
+        if isinstance(data, dict):
+            up_header['Content-Type'] = 'application/json'
+            data = json.dumps(data)
+        else:
+            up_header['Content-Type'] = 'application/octet-stream'
+        resp = requests.put(
+            req_url, verify=self._verify_ssl,
+            headers=up_header,
+            data=data)
+        resp.raise_for_status()
+        if len(resp.content) > 0:
+            return resp.json()
+
     def get_stock(self) -> List[CurrentStockResponse]:
-        req_url = urljoin(self._base_url, "stock")
-        resp = requests.get(req_url, verify=self._verify_ssl, headers=self._headers)
-        if resp.status_code != 200 or not resp.text:
-            return
-        parsed_json = resp.json()
+        parsed_json = self._do_get_request("stock")
         return [CurrentStockResponse(response) for response in parsed_json]
 
     def get_volatile_stock(self) -> CurrentVolatilStockResponse:
-        req_url = urljoin(self._base_url, "stock/volatile")
-        resp = requests.get(req_url, verify=self._verify_ssl, headers=self._headers)
-        parsed_json = resp.json()
+        parsed_json = self._do_get_request("stock/volatile")
         return CurrentVolatilStockResponse(parsed_json)
 
     def get_product(self, product_id) -> ProductDetailsResponse:
-        req_url = urljoin(urljoin(self._base_url, "stock/products/"), str(product_id))
-        resp = requests.get(req_url, verify=self._verify_ssl, headers=self._headers)
-        if resp.status_code != 200 or not resp.text:
-            return
-        parsed_json = resp.json()
-        return ProductDetailsResponse(parsed_json)
+        url = f"stock/products/{product_id}"
+        parsed_json = self._do_get_request(url)
+        if parsed_json:
+            return ProductDetailsResponse(parsed_json)
 
     def get_chores(self) -> List[CurrentChoreResponse]:
-        req_url = urljoin(self._base_url, "chores")
-        resp = requests.get(req_url, verify=self._verify_ssl, headers=self._headers)
-        parsed_json = resp.json()
+        parsed_json = self._do_get_request("chores")
         return [CurrentChoreResponse(chore) for chore in parsed_json]
 
     def get_chore(self, chore_id: int) -> ChoreDetailsResponse:
-        req_url = urljoin(urljoin(self._base_url, "chores/"), str(chore_id))
-        resp = requests.get(req_url, verify=self._verify_ssl, headers=self._headers)
-        parsed_json = resp.json()
-        return ChoreDetailsResponse(parsed_json)
+        url = f"chores/{chore_id}"
+        parsed_json = self._do_get_request(url)
+        if parsed_json:
+            return ChoreDetailsResponse(parsed_json)
 
     def execute_chore(self, chore_id: int, done_by: int = None, tracked_time: datetime = datetime.now()):
         # Grocy API expects UTC time; time returned from datetime.now() is local time without timezone
@@ -374,8 +439,7 @@ class GrocyApiClient(object):
         if done_by is not None:
             data["done_by"] = done_by
 
-        req_url = urljoin(urljoin(urljoin(self._base_url, "chores/"), str(chore_id) + "/"), "execute")
-        requests.post(req_url, verify=self._verify_ssl, headers=self._headers, data=data)
+        return self._do_post_request(f"chores/{chore_id}/execute", data)
 
     def add_product(self, product_id, amount: float, price: float, best_before_date: datetime = None,
                     transaction_type: TransactionType = TransactionType.PURCHASE):
@@ -388,8 +452,7 @@ class GrocyApiClient(object):
         if best_before_date is not None:
             data["best_before_date"] = best_before_date.strftime('%Y-%m-%d')
 
-        req_url = urljoin(urljoin(urljoin(self._base_url, "stock/products/"), str(product_id) + "/"), "add")
-        requests.post(req_url, verify=self._verify_ssl, headers=self._headers, data=data)
+        return self._do_post_request(f"stock/products/{product_id}/add", data)
 
     def consume_product(self, product_id: int, amount: float = 1, spoiled: bool = False,
                         transaction_type: TransactionType = TransactionType.CONSUME):
@@ -399,104 +462,78 @@ class GrocyApiClient(object):
             "transaction_type": transaction_type.value
         }
 
-        req_url = urljoin(urljoin(urljoin(self._base_url, "stock/products/"), str(product_id) + "/"), "consume")
-        requests.post(req_url, verify=self._verify_ssl, headers=self._headers, data=data)
+        self._do_post_request(f"stock/products/{product_id}/consume", data)
 
-        
     def get_shopping_list(self) -> List[ShoppingListItem]:
-        req_url = urljoin(self._base_url, "objects/shopping_list")
-        resp = requests.get(req_url, verify=self._verify_ssl, headers=self._headers)
-        if resp.status_code != 200:
-            return
-        parsed_json = resp.json()
+        parsed_json = self._do_get_request("objects/shopping_list")
         return [ShoppingListItem(response) for response in parsed_json]
 
-    def add_missing_product_to_shopping_list(self, shopping_list_id: int = 1):
-        data = {
-            "list_id": shopping_list_id
-        }
+    def add_missing_product_to_shopping_list(self, shopping_list_id: int = None):
+        data = None
+        if shopping_list_id:
+            data = {
+                "list_id": shopping_list_id
+            }
+
+        self._do_post_request("stock/shoppinglist/add-missing-products", data)
 
-        req_url = urljoin(self._base_url, "stock/shoppinglist/add-missing-products")
-        resp = requests.post(req_url, verify=self._verify_ssl, headers=self._headers, data=data)
-        return resp
-    
     def add_product_to_shopping_list(self, product_id: int, shopping_list_id: int = 1, amount: int = 1):
         data = {
             "product_id": product_id,
             "list_id": shopping_list_id,
             "product_amount": amount
         }
+        self._do_post_request("stock/shoppinglist/add-product", data)
 
-        req_url = urljoin(self._base_url, "stock/shoppinglist/add-product")
-        resp = requests.post(req_url, verify=self._verify_ssl, headers=self._headers, data=data)
-        return resp
-    
     def clear_shopping_list(self, shopping_list_id: int = 1):
         data = {
             "list_id": shopping_list_id
         }
 
-        req_url = urljoin(self._base_url, "stock/shoppinglist/clear")
-        resp = requests.post(req_url, verify=self._verify_ssl, headers=self._headers, data=data)
-        return resp
-            
+        self._do_post_request("stock/shoppinglist/clear", data)
+
     def remove_product_in_shopping_list(self, product_id: int, shopping_list_id: int = 1, amount: int = 1):
         data = {
             "product_id": product_id,
             "list_id": shopping_list_id,
             "product_amount": amount
         }
-        req_url = urljoin(self._base_url, "stock/shoppinglist/remove-product")
-        resp = requests.post(req_url, verify=self._verify_ssl, headers=self._headers, data=data)
-        return resp
-        
+        self._do_post_request("stock/shoppinglist/remove-product", data)
+
     def get_product_groups(self) -> List[LocationData]:
-        req_url = urljoin(self._base_url, "objects/product_groups")
-        resp = requests.get(req_url, verify=self._verify_ssl, headers=self._headers)
-        if resp.status_code != 200 or not resp.text:
-            return
-        parsed_json = resp.json()
+        parsed_json = self._do_get_request("objects/product_groups")
         return [LocationData(response) for response in parsed_json]
-        
+
     def upload_product_picture(self, product_id: int, pic_path: str):
-        if not os.path.exists(pic_path):
-            return
-        up_header = self._headers.copy()
-        up_header['accept'] = '*/*'
-        up_header['Content-Type'] = 'application/octet-stream'
         b64fn = base64.b64encode('{}.jpg'.format(product_id).encode('ascii'))
-        req_url = urljoin(urljoin(self._base_url, "files/productpictures/" ), str(b64fn, "utf-8"))
-        with open(pic_path,'rb') as pic:
-            resp =requests.put(req_url, verify=self._verify_ssl, headers=up_header , data=pic)
-            return resp
-            
+        req_url = "files/productpictures/" + str(b64fn, "utf-8")
+        with open(pic_path, 'rb') as pic:
+            self._do_put_request(req_url, pic)
+
     def update_product_pic(self, product_id: int):
-        pic_name = '{}.jpg'.format(product_id)
-        data = { "picture_file_name":  pic_name }
-        up_header = self._headers.copy()
-        up_header['accept'] = '*/*'
-        up_header['Content-Type'] = 'application/json'
-        req_url = urljoin(urljoin(self._base_url, "objects/products/"), str(product_id))
-        resp = requests.put(req_url, verify=self._verify_ssl, headers=up_header , data=json.dumps(data))
-        return resp
-            
+        pic_name = f"{product_id}.jpg"
+        data = {"picture_file_name": pic_name}
+        self._do_put_request(f"objects/products/{product_id}", data)
+
     def get_userfields(self, entity: str, object_id: int):
-        req_url = urljoin(urljoin(urljoin(self._base_url, "userfields/"), entity + "/"), str(object_id))
-        resp = requests.get(req_url, verify=self._verify_ssl, headers=self._headers)
-        if not resp:
-            return resp
-        return resp.json()
-        
+        url = f"userfields/{entity}/{object_id}"
+        return self._do_get_request(url)
+
     def set_userfields(self, entity: str, object_id: int, key: str, value):
         data = {
             key: value
         }
-        req_url = urljoin(urljoin(urljoin(self._base_url, "userfields/"), entity + "/"), str(object_id))
-        resp = requests.put(req_url, verify=self._verify_ssl, headers=self._headers, data=data)
-        return resp
+        self._do_put_request(f"userfields/{entity}/{object_id}", data)
 
     def get_last_db_changed(self):
-        req_url = urljoin(self._base_url, "system/db-changed-time")
-        resp = requests.get(req_url, verify=self._verify_ssl, headers=self._headers)
-        last_change_timestamp = parse_date(resp.json().get('changed_time'))
+        resp = self._do_get_request("system/db-changed-time")
+        last_change_timestamp = parse_date(resp.get('changed_time'))
         return last_change_timestamp
+
+    def get_tasks(self) -> List[TaskResponse]:
+        parsed_json = self._do_get_request("tasks")
+        return [TaskResponse(data) for data in parsed_json]
+
+    def complete_task(self, task_id: int):
+        url = f"tasks/{task_id}/complete"
+        self._do_post_request(url)
